@@ -1,6 +1,6 @@
 # Orca
 #
-# Copyright 2004-2006 Sun Microsystems Inc.
+# Copyright 2004-2007 Sun Microsystems Inc.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Library General Public
@@ -19,10 +19,10 @@
 
 """Provides the FocusTrackingPresenter for Orca."""
 
-__id__        = "$Id: focus_tracking_presenter.py 1825 2006-12-13 22:35:06Z wwalker $"
-__version__   = "$Revision: 1825 $"
-__date__      = "$Date: 2006-12-13 17:35:06 -0500 (Wed, 13 Dec 2006) $"
-__copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
+__id__        = "$Id: focus_tracking_presenter.py 2064 2007-02-27 16:46:00Z wwalker $"
+__version__   = "$Revision: 2064 $"
+__date__      = "$Date: 2007-02-27 11:46:00 -0500 (Tue, 27 Feb 2007) $"
+__copyright__ = "Copyright (c) 2005-2007 Sun Microsystems Inc."
 __license__   = "LGPL"
 
 import gobject
@@ -419,7 +419,10 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                 atspi.Accessible.deleteAccessible(event.source)
                 return
 
-            debug.printDetails(debug.LEVEL_FINEST, "    ", event.source)
+            if (not debug.eventDebugFilter) \
+                or (debug.eventDebugFilter \
+                    and debug.eventDebugFilter.match(event.type)):
+                debug.printDetails(debug.LEVEL_FINEST, "    ", event.source)
 
         except CORBA.COMM_FAILURE:
             debug.printException(debug.LEVEL_WARNING)
@@ -453,7 +456,15 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         #
         retryCount = 0
         oldLocusOfFocus = orca_state.locusOfFocus
-        while retryCount <= settings.commFailureAttemptLimit:
+        try:
+            s = self._getScript(event.source.app)
+        except:
+            s = None
+            debug.printException(debug.LEVEL_WARNING)
+            debug.println(debug.LEVEL_WARNING,
+                          "ERROR: received an event, but Script is None")
+
+        while s and retryCount <= s.commFailureAttemptLimit:
             try:
                 if not event.source.state.count( \
                                       atspi.Accessibility.STATE_ICONIFIED):
@@ -496,7 +507,6 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                         braille.setupKeyRanges(\
                             orca_state.activeScript.brailleBindings.keys())
 
-                    s = self._getScript(event.source.app)
                     s.processObjectEvent(event)
                     if retryCount:
                         debug.println(debug.LEVEL_WARNING,
@@ -508,7 +518,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                               "COMM_FAILURE above while processing: " \
                               + event.type)
                 retryCount += 1
-                if retryCount <= settings.commFailureAttemptLimit:
+                if retryCount <= s.commFailureAttemptLimit:
                     # We want the old locus of focus to be reset so
                     # the proper stuff will be spoken if the locus
                     # of focus changed during our last attempt at
@@ -517,7 +527,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     orca_state.locusOfFocus = oldLocusOfFocus
                     debug.println(debug.LEVEL_WARNING,
                                   "  TRYING AGAIN (%d)" % retryCount)
-                    time.sleep(settings.commFailureWaitTime)
+                    time.sleep(s.commFailureWaitTime)
                 else:
                     debug.println(debug.LEVEL_WARNING,
                                   "  GIVING UP AFTER %d TRIES" \
@@ -601,9 +611,22 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             # we want it to ref, thus allowing things to survive until
             # they are processed on the gidle thread.
             #
-            debug.println(debug.LEVEL_FINEST,
-                          "---------> QUEUEING EVENT %s" % e.type)
+            # If the event doesn't have a source or that source is not marked
+            # valid, then we don't care about this event. Just return.
+            #
             event = atspi.Event(e)
+            if not event.source or not event.source.valid:
+                debug.println(debug.LEVEL_FINEST,
+                      "---------> IGNORING INVALID EVENT %s" % e.type)
+                if settings.debugEventQueue:
+                    self._enqueueEventCount -= 1
+                return
+
+            if (not debug.eventDebugFilter) \
+                or (debug.eventDebugFilter \
+                    and debug.eventDebugFilter.match(e.type)):
+                debug.println(debug.LEVEL_FINEST,
+                              "---------> QUEUEING EVENT %s" % e.type)
 
         if event:
             if settings.debugEventQueue:
@@ -695,12 +718,12 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                               "\n^^^^^ PROCESS BRAILLE EVENT %d ^^^^^"\
                               % event.event)
             else:
-                debug.println(debug.LEVEL_FINEST, "DEQUEUED EVENT %s <----------" \
-                              % event.type)
-
                 if (not debug.eventDebugFilter) \
                     or (debug.eventDebugFilter \
                         and debug.eventDebugFilter.match(event.type)):
+                    debug.println(debug.LEVEL_FINEST,
+                                  "DEQUEUED EVENT %s <----------" \
+                                  % event.type)
                     debug.println(debug.eventDebugLevel,
                                   "\nvvvvv PROCESS OBJECT EVENT %s vvvvv" \
                                   % event.type)
