@@ -19,9 +19,9 @@
 
 """Custom script for StarOffice and OpenOffice."""
 
-__id__        = "$Id: StarOffice.py,v 1.73 2006/12/14 16:48:08 richb Exp $"
-__version__   = "$Revision: 1.73 $"
-__date__      = "$Date: 2006/12/14 16:48:08 $"
+__id__        = "$Id: StarOffice.py 2068 2007-02-28 15:24:41Z richb $"
+__version__   = "$Revision: 2068 $"
+__date__      = "$Date: 2007-02-28 10:24:41 -0500 (Wed, 28 Feb 2007) $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -50,7 +50,24 @@ inputLineForCell = None
 dynamicColumnHeaders = {}
 dynamicRowHeaders = {}
 
-def getCalcTable(obj):
+def adjustForWriterTable(obj):
+    """Check to see if we are in Writer, where the object with focus is a
+    paragraph, and the parent is the table cell. If it is, then, return the
+    parent table cell otherwise return the current object.
+
+    Arguments:
+    - obj: the accessible object to check.
+
+    Returns parent table cell (if in a Writer table ) or the current object.
+    """
+
+    if obj.role == rolenames.ROLE_PARAGRAPH and \
+       obj.parent.role == rolenames.ROLE_TABLE_CELL:
+        return obj.parent
+    else:
+        return obj
+
+def getTable(obj):
     """Get the table that this table cell is in.
 
     Arguments:
@@ -61,6 +78,7 @@ def getCalcTable(obj):
     """
 
     table = None
+    obj = adjustForWriterTable(obj)
     if obj.role == rolenames.ROLE_TABLE_CELL and obj.parent:
         table = obj.parent.table
 
@@ -77,6 +95,7 @@ def getDynamicColumnHeaderCell(obj, column):
     Return the dynamic column header cell associated with the given table cell.
     """
 
+    obj = adjustForWriterTable(obj)
     accCell = None
     parent = obj.parent
     if parent and parent.table:
@@ -97,6 +116,7 @@ def getDynamicRowHeaderCell(obj, row):
     Return the dynamic row header cell associated with the given table cell.
     """
 
+    obj = adjustForWriterTable(obj)
     accCell = None
     parent = obj.parent
     if parent and parent.table:
@@ -196,36 +216,48 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
 
         regions = []
 
+        # Check to see if this spread sheet cell has either a dynamic
+        # column heading or row heading (or both) associated with it.
+        # If it does, then braille those first before brailling the
+        # cell contents.
+        #
+        table = getTable(obj)
+        parent = obj.parent
+
+        if parent.__dict__.has_key("lastColumn") and \
+           parent.lastColumn != parent.table.getColumnAtIndex(obj.index):
+            if dynamicColumnHeaders.has_key(table):
+                row = dynamicColumnHeaders[table]
+                header = getDynamicRowHeaderCell(obj, row)
+                if header.childCount > 0:
+                    for i in range(0, header.childCount):
+                        child = header.child(i)
+                        text = self._script.getText(child, 0, -1)
+                        if text:
+                            regions.append(braille.Region(" " + text + " "))
+                elif header.text:
+                    text = self._script.getText(header, 0, -1)
+                    if text:
+                        regions.append(braille.Region(" " + text + " "))
+
+        if parent.__dict__.has_key("lastRow") and \
+           parent.lastRow != parent.table.getRowAtIndex(obj.index):
+            if dynamicRowHeaders.has_key(table):
+                column = dynamicRowHeaders[table]
+                header = getDynamicColumnHeaderCell(obj, column)
+                if header.childCount > 0:
+                    for i in range(0, header.childCount):
+                        child = header.child(i)
+                        text = self._script.getText(child, 0, -1)
+                        if text:
+                            regions.append(braille.Region(" " + text + " "))
+                elif header.text:
+                    text = self._script.getText(header, 0, -1)
+                    if text:
+                        regions.append(braille.Region(" " + text + " "))
+
         if isSpreadSheetCell(obj):
 
-            # Check to see if this spread sheet cell has either a dynamic
-            # column heading or row heading (or both) associated with it.
-            # If it does, then braille those first before brailling the
-            # cell contents.
-            #
-            table = getCalcTable(obj)
-            parent = obj.parent
-
-            if parent.__dict__.has_key("lastColumn") and \
-               parent.lastColumn != parent.table.getColumnAtIndex(obj.index):
-                if dynamicColumnHeaders.has_key(table):
-                    row = dynamicColumnHeaders[table]
-                    header = getDynamicRowHeaderCell(obj, row)
-                    if header.text:
-                        text = self._script.getText(header, 0, -1)
-                        if text:
-                            regions.append(braille.Region(" " + text + " "))
-
-            if parent.__dict__.has_key("lastRow") and \
-               parent.lastRow != parent.table.getRowAtIndex(obj.index):
-                if dynamicRowHeaders.has_key(table):
-                    column = dynamicRowHeaders[table]
-                    header = getDynamicColumnHeaderCell(obj, column)
-                    if header.text:
-                        text = self._script.getText(header, 0, -1)
-                        if text:
-                            regions.append(braille.Region(" " + text + " "))
-
             # Adding in a check here to make sure that the parent is a
             # valid table. It's possible that the parent could be a
             # table cell too (see bug #351501).
@@ -280,7 +312,10 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
             regions = [regions, focusRegion]
         else:
             brailleGen = braillegenerator.BrailleGenerator
-            regions = brailleGen._getBrailleRegionsForTableCellRow(self, obj)
+            [cellRegions, focusRegion] = \
+                brailleGen._getBrailleRegionsForTableCellRow(self, obj)
+            regions.extend(cellRegions)
+            regions = [regions, focusRegion]
 
         return regions
 
@@ -350,6 +385,11 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
     Overrides _getSpeechForTableCell so that, when we are in a spread
     sheet, we can speak the location of the table cell as well as the
     contents.
+    Overrides _getSpeechForToggleButton so that, when the toggle buttons
+    on the Formatting toolbar change state, we provide both the name and
+    the state (as "on" or "off")
+    Overrides _getSpeechForPushButton because sometimes the toggle buttons
+    on the Formatting toolbar claim to be push buttons.
     """
     def __init__(self, script):
         speechgenerator.SpeechGenerator.__init__(self, script)
@@ -411,38 +451,51 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
         utterances = []
 
+        if not already_focused:
+
+            # Check to see if this spread sheet cell has either a dynamic
+            # column heading or row heading (or both) associated with it.
+            # If it does, then speak those first before speaking the cell
+            # contents.
+            #
+            table = getTable(obj)
+            parent = obj.parent
+
+            if parent.__dict__.has_key("lastColumn") and \
+               parent.lastColumn != \
+               parent.table.getColumnAtIndex(obj.index):
+                if dynamicColumnHeaders.has_key(table):
+                    row = dynamicColumnHeaders[table]
+                    header = getDynamicRowHeaderCell(obj, row)
+                    if header.childCount > 0:
+                        for i in range(0, header.childCount):
+                            child = header.child(i)
+                            text = self._script.getText(child, 0, -1)
+                            if text:
+                                utterances.append(text)
+                    elif header.text:
+                        text = self._script.getText(header, 0, -1)
+                        if text:
+                            utterances.append(text)
+
+            if parent.__dict__.has_key("lastRow") and \
+               parent.lastRow != parent.table.getRowAtIndex(obj.index):
+                if dynamicRowHeaders.has_key(table):
+                    column = dynamicRowHeaders[table]
+                    header = getDynamicColumnHeaderCell(obj, column)
+                    if header.childCount > 0:
+                        for i in range(0, header.childCount):
+                            child = header.child(i)
+                            text = self._script.getText(child, 0, -1)
+                            if text:
+                                utterances.append(text)
+                    elif header.text:
+                        text = self._script.getText(header, 0, -1)
+                        if text:
+                            utterances.append(text)
+
         if isSpreadSheetCell(obj):
-            if (not already_focused):
-
-                # Check to see if this spread sheet cell has either a dynamic
-                # column heading or row heading (or both) associated with it.
-                # If it does, then speak those first before speaking the cell
-                # contents.
-                #
-                table = getCalcTable(obj)
-                parent = obj.parent
-
-                if parent.__dict__.has_key("lastColumn") and \
-                   parent.lastColumn != \
-                   parent.table.getColumnAtIndex(obj.index):
-                    if dynamicColumnHeaders.has_key(table):
-                        row = dynamicColumnHeaders[table]
-                        header = getDynamicRowHeaderCell(obj, row)
-                        if header.text:
-                            text = self._script.getText(header, 0, -1)
-                            if text:
-                                utterances.append(text)
-
-                if parent.__dict__.has_key("lastRow") and \
-                   parent.lastRow != parent.table.getRowAtIndex(obj.index):
-                    if dynamicRowHeaders.has_key(table):
-                        column = dynamicRowHeaders[table]
-                        header = getDynamicColumnHeaderCell(obj, column)
-                        if header.text:
-                            text = self._script.getText(header, 0, -1)
-                            if text:
-                                utterances.append(text)
-
+            if not already_focused:
                 if settings.readTableCellRow:
                     parent = obj.parent
                     row = parent.table.getRowAtIndex(obj.index)
@@ -477,8 +530,8 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
                                                              already_focused))
         else:
             speechGen = speechgenerator.SpeechGenerator
-            utterances = speechGen._getSpeechForTableCellRow(self, obj,
-                                                             already_focused)
+            utterances.extend(speechGen._getSpeechForTableCellRow(self, obj,
+                                                             already_focused))
 
         return utterances
 
@@ -529,6 +582,60 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
         return utterances
 
+    def _getSpeechForToggleButton(self, obj, already_focused):
+        """Get the speech for a toggle button.  We always want to speak the
+        state if it's on a toolbar.
+
+        Arguments:
+        - obj: the toggle button
+        - already_focused: False if object just received focus
+
+        Returns a list of utterances to be spoken for the object.
+        """
+
+        utterances = []
+        if obj.parent.role == rolenames.ROLE_TOOL_BAR:
+            if obj.state.count(atspi.Accessibility.STATE_CHECKED):
+                checkedState = _("on")
+            else:
+                checkedState = _("off")
+
+            utterances.append(obj.name)
+            utterances.append(checkedState)
+        else:
+            speechGen = speechgenerator.SpeechGenerator
+            utterances.extend(speechGen._getSpeechForToggleButton(self, obj,
+                                                             already_focused))
+
+        return utterances
+
+    def _getSpeechForPushButton(self, obj, already_focused):
+        """Get the speech for a push button.  We always want to speak the
+        state if it's on a toolbar.
+
+        Arguments:
+        - obj: the push button
+        - already_focused: False if object just received focus
+
+        Returns a list of utterances to be spoken for the object.
+        """
+
+        utterances = []
+        if obj.parent.role == rolenames.ROLE_TOOL_BAR:
+            if obj.state.count(atspi.Accessibility.STATE_CHECKED):
+                checkedState = _("on")
+            else:
+                checkedState = _("off")
+
+            utterances.append(obj.name)
+            utterances.append(checkedState)
+        else:
+            speechGen = speechgenerator.SpeechGenerator
+            utterances.extend(speechGen._getSpeechForPushButton(self, obj,
+                                                             already_focused))
+
+        return utterances
+
 ########################################################################
 #                                                                      #
 # The StarOffice script class.                                         #
@@ -572,6 +679,24 @@ class Script(default.Script):
         #
         self.currentParagraph = None
 
+        # Set the number of retries after a COMM_FAILURE to 1. The default
+        # of 5 was not allowing Orca to be responsive in the event of OOo
+        # going into crash recovery mode (see bug #397787).
+        #
+        self.commFailureAttemptLimit = 1
+
+        # The default set of text attributes to speak to the user. The
+        # only difference over the default set in settings.py is to add
+        # in "left-margin:" and "right-margin:".
+
+        settings.enabledTextAttributes = "size:; family-name:; weight:400; indent:0; left-margin:0; right-margin:0; underline:none; strikethrough:false; justification:left; style:normal;"
+
+        # [[[TODO: JD - HACK because we won't get events from toggle
+        # buttons on the Formatting toolbar until we "tickle/poke"
+        # the hierarchy. But we only want to do it once.  
+        # See bug #363830 and OOo issue #70872.]]]
+        #
+        self.tickled = None
 
     def getBrailleGenerator(self):
         """Returns the braille generator for this script.
@@ -701,7 +826,7 @@ class Script(default.Script):
                         "StarOffice.speakInputLine: contents: %s" % inputLine)
                 speech.speak(inputLine)
 
-    def getCalcRow(self, cell):
+    def getTableRow(self, cell):
         """Get the row number in the table that this table cell is on.
 
         Arguments:
@@ -712,6 +837,7 @@ class Script(default.Script):
         """
 
         row = None
+        cell = adjustForWriterTable(cell)
         if cell.role == rolenames.ROLE_TABLE_CELL:
             parent = cell.parent
             if parent and parent.table:
@@ -719,7 +845,7 @@ class Script(default.Script):
 
         return row
 
-    def getCalcColumn(self, cell):
+    def getTableColumn(self, cell):
         """Get the column number in the table that this table cell is on.
 
         Arguments:
@@ -730,6 +856,7 @@ class Script(default.Script):
         """
 
         column = None
+        cell = adjustForWriterTable(cell)
         if cell.role == rolenames.ROLE_TABLE_CELL:
             parent = cell.parent
             if parent and parent.table:
@@ -752,14 +879,14 @@ class Script(default.Script):
         - inputEvent: if not None, the input event that caused this action.
         """
 
-        global dynamicColumnHeaders, getCalcTable
+        global dynamicColumnHeaders, getTable
 
         debug.println(self.debugLevel, "StarOffice.setDynamicColumnHeaders.")
 
         clickCount = util.getClickCount(self.lastDynamicEvent, inputEvent)
-        table = getCalcTable(orca_state.locusOfFocus)
+        table = getTable(orca_state.locusOfFocus)
         if table:
-            row = self.getCalcRow(orca_state.locusOfFocus)
+            row = self.getTableRow(orca_state.locusOfFocus)
             if clickCount == 2:
                 try:
                     del dynamicColumnHeaders[table]
@@ -815,14 +942,14 @@ class Script(default.Script):
         - inputEvent: if not None, the input event that caused this action.
         """
 
-        global dynamicRowHeaders, getCalcTable
+        global dynamicRowHeaders, getTable
 
         debug.println(self.debugLevel, "StarOffice.setDynamicRowHeaders.")
 
         clickCount = util.getClickCount(self.lastDynamicEvent, inputEvent)
-        table = getCalcTable(orca_state.locusOfFocus)
+        table = getTable(orca_state.locusOfFocus)
         if table:
-            column = self.getCalcColumn(orca_state.locusOfFocus)
+            column = self.getTableColumn(orca_state.locusOfFocus)
             if clickCount == 2:
                 try:
                     del dynamicRowHeaders[table]
@@ -862,6 +989,7 @@ class Script(default.Script):
 
         textLength = paragraph[0].text.characterCount
         startFound = False
+        startOff = 0
         endOff = textLength
         for i in range(0, textLength):
             attributes = paragraph[0].text.getAttributes(i)
@@ -1424,13 +1552,49 @@ class Script(default.Script):
             if self.findCommandRun:
                 return
 
+            # [[[TODO: JD - HACK because we won't get events from toggle
+            # buttons on the Formatting toolbar until we "tickle/poke"
+            # the hierarchy. But we only want to do it once.
+            # See bug #363830 and OOo issue #70872.]]]
+            #
+            if not self.tickled:
+                self.tickled = self.flatReviewContextClass(self)
+
+        # Announce when the toolbar buttons are toggled if we just toggled
+        # them; not if we navigated to some text.
+        #
+        if event.type == "object:state-changed:checked" and \
+           (event.source.role == rolenames.ROLE_TOGGLE_BUTTON or \
+            event.source.role == rolenames.ROLE_PUSH_BUTTON):
+            weToggledIt = False
+            if isinstance(orca_state.lastInputEvent, \
+                          input_event.MouseButtonEvent):
+                x = orca_state.lastInputEvent.x
+                y = orca_state.lastInputEvent.y
+                weToggledIt = event.source.component.contains(x, y, 0)
+
+            elif isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
+                keyString = orca_state.lastInputEvent.event_string  
+                navKeys = ["Up", "Down", "Page_Up", "Page_Down", "Home", "End"]
+                wasCommand = orca_state.lastInputEvent.modifiers \
+                             & (1 << atspi.Accessibility.MODIFIER_CONTROL \
+                              | 1 << atspi.Accessibility.MODIFIER_ALT \
+                              | 1 << atspi.Accessibility.MODIFIER_META \
+                              | 1 << atspi.Accessibility.MODIFIER_META2 \
+                              | 1 << atspi.Accessibility.MODIFIER_META3)
+                weToggledIt = wasCommand and keyString not in navKeys
+
+            if weToggledIt:
+                speech.speakUtterances(self.speechGenerator.getSpeech( \
+                                       event.source, False))
+
         # If we are FOCUSED on a paragraph inside a table cell (in Writer),
         # then speak/braille that parent table cell (see bug #382415).
         #
         if event.type == "object:state-changed:focused" and \
            event.source.role == rolenames.ROLE_PARAGRAPH and \
            event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
-           event.detail1 == 1 and \
+           event.detail1 == 0 and \
            event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
 
             # Check to see if the last input event was "Up" or "Down".
@@ -1439,7 +1603,9 @@ class Script(default.Script):
             # get the speech for that single child, otherwise speak/braille
             # the parent table cell.
             #
-            event_string = orca_state.lastInputEvent.event_string
+            event_string = None
+            if isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
+                event_string = orca_state.lastInputEvent.event_string
             if (event_string == "Up" or event_string == "Down") and \
                event.source.parent == self.lastCell and \
                event.source.parent.childCount > 1:
